package org.rlcommunity.environments.mountaincar;

import java.net.URL;
import java.util.StringTokenizer;
import java.util.Vector;

import org.rlcommunity.environments.mountaincar.messages.MCGoalResponse;
import org.rlcommunity.environments.mountaincar.messages.MCHeightResponse;
import org.rlcommunity.environments.mountaincar.messages.MCStateResponse;
import rlVizLib.Environments.EnvironmentBase;
import rlVizLib.general.ParameterHolder;
import rlVizLib.messaging.NotAnRLVizMessageException;
import rlVizLib.messaging.environment.EnvironmentMessageParser;
import rlVizLib.messaging.environment.EnvironmentMessages;
import rlVizLib.messaging.interfaces.HasAVisualizerInterface;
import rlVizLib.messaging.interfaces.getEnvMaxMinsInterface;
import rlVizLib.messaging.interfaces.getEnvObsForStateInterface;
import org.rlcommunity.rlglue.codec.types.Action;
import org.rlcommunity.rlglue.codec.types.Observation;

import org.rlcommunity.rlglue.codec.types.Reward_observation_terminal;

import java.util.Random;
import org.rlcommunity.environments.mountaincar.visualizer.PlayroomVisualizer;
import org.rlcommunity.rlglue.codec.taskspec.TaskSpec;
import org.rlcommunity.rlglue.codec.taskspec.TaskSpecVRLGLUE3;
import org.rlcommunity.rlglue.codec.taskspec.ranges.DoubleRange;
import org.rlcommunity.rlglue.codec.taskspec.ranges.IntRange;
import org.rlcommunity.rlglue.codec.util.EnvironmentLoader;
import rlVizLib.general.hasVersionDetails;

/*
 * April 2010
 * This is the Java Version Playroom Domain.  
 * @author Nitish Srivastava
 */
import rlVizLib.messaging.environmentShell.TaskSpecPayload;
import rlVizLib.messaging.interfaces.HasImageInterface;

public class Playroom extends EnvironmentBase implements
//        getEnvMaxMinsInterface,
//        getEnvObsForStateInterface,
        HasImageInterface {

    static final int numActions = 9;
    static int count;
    protected PlayroomState theState = null;
    //through the messaging system and RL-Viz.
    protected Vector<PlayroomState> savedStates = null;
    //Problem parameters have been moved to Playroom State
    private Random randomGenerator = new Random();

    public static TaskSpecPayload getTaskSpecPayload(ParameterHolder P) {
        Playroom theMC = new Playroom(P);
        String taskSpec = theMC.makeTaskSpec();
        return new TaskSpecPayload(taskSpec, false, "");
    }

    private String makeTaskSpec() {
        TaskSpecVRLGLUE3 theTaskSpecObject = new TaskSpecVRLGLUE3();
        theTaskSpecObject.setEpisodic();
        theTaskSpecObject.setDiscountFactor(1.0d);
        theTaskSpecObject.addDiscreteObservation(new IntRange(0,6,3));//0 for unobservable
	theTaskSpecObject.addDiscreteObservation(new IntRange(0,1,4));
        theTaskSpecObject.addDiscreteAction(new IntRange(0,9));
        theTaskSpecObject.setRewardRange(new DoubleRange(-1, 10));
        theTaskSpecObject.setExtra("EnvName:Playroom Revision:" + this.getClass().getPackage().getImplementationVersion());
        String taskSpecString = theTaskSpecObject.toTaskSpec();
        TaskSpec.checkTaskSpec(taskSpecString);
        return taskSpecString;

    }
    public String env_init() {
        savedStates = new Vector<PlayroomState>();
        return makeTaskSpec();
   }

    /**
     * Restart the car on the mountain.  
     * randomStarts is set.
     * @return
     */
    public Observation env_start() {

        for (int i=0;i<theState.numobjects;i++){
		theState.position[i][0] = theState.minpos+randomGenerator.nextInt(theState.maxpos-theState.minpos+1);
		theState.position[i][1] = theState.minpos+randomGenerator.nextInt(-theState.minpos+theState.maxpos+1);
	}

	/*theState.position[1][0]=theState.position[0][0];
	theState.position[1][1]=theState.position[0][1];
	*/
	theState.isLightOn=false;
	theState.isMusicOn=false;
	theState.isMonkeyCrying=false;
	theState.isBellRinging=false;
	theState.displayState();
	count=0;
        return makeObservation();
    }

    /**
     * Takes a step.
     * @param theAction
     * @return
     */
    public Reward_observation_terminal env_step(Action theAction) {
        //System.out.print(" "+count++ + " ");
	int a = theAction.intArray[0];
        theState.isBellRinging=false;	//Bell rings only once
	theState.update(a);
	//theState.displayState();
	Reward_observation_terminal ro= new Reward_observation_terminal();
	ro.terminal=(theState.inGoalState()?1:0);
	ro.o=makeObservation();
	ro.r=theState.getReward();
        return ro;
    }

    /**
     * Return the ParameterHolder object that contains the default parameters for
     * mountain car.  The only parameter is random start states.
     * @return
     */
    public static ParameterHolder getDefaultParameters() {
        ParameterHolder p = new ParameterHolder();
        rlVizLib.utilities.UtilityShop.setVersionDetails(p, new DetailsProvider());
        p.addBooleanParam("randomStartStates", true);
        return p;
    }

    /**
     * Create a new mountain car environment using parameter settings in p.
     * @param p
     */
    public Playroom(ParameterHolder p) {
        super();
        theState = new PlayroomState(randomGenerator);
        if (p != null) {
            if (!p.isNull()) {
                theState.randomStarts = p.getBooleanParam("randomStartStates");
            }
        }
    }

    public Playroom() {
        this(getDefaultParameters());
    }

    /**
     * Handles messages that find out the version, what visualizer is available, 
     * etc.
     * @param theMessage
     * @return
     */
    public String env_message(String theMessage) {
        EnvironmentMessages theMessageObject;
        try {
            theMessageObject = EnvironmentMessageParser.parseMessage(theMessage);
        } catch (NotAnRLVizMessageException e) {
            System.err.println("Someone sent Playroom a message that wasn't RL-Viz compatible");
            return "I only respond to RL-Viz messages!";
        }

        if (theMessageObject.canHandleAutomatically(this)) {
            String theResponseString = theMessageObject.handleAutomatically(this);
            return theResponseString;
        }

        //If it wasn't handled automatically, maybe its a custom Mountain Car Message
        /*if (theMessageObject.getTheMessageType() == rlVizLib.messaging.environment.EnvMessageType.kEnvCustom.id()) {

            String theCustomType = theMessageObject.getPayLoad();

            if (theCustomType.equals("GETMCSTATE")) {
                //It is a request for the state
                double position = theState.position;
                double velocity = theState.velocity;
                double height = this.getHeight();
                double deltaheight = theState.getHeightAtPosition(position + .05);
                MCStateResponse theResponseObject = new MCStateResponse(position, velocity, height, deltaheight);
                return theResponseObject.makeStringResponse();
            }

            if (theCustomType.startsWith("GETHEIGHTS")) {
                Vector<Double> theHeights = new Vector<Double>();

                StringTokenizer theTokenizer = new StringTokenizer(theCustomType, ":");
                //throw away the first token
                theTokenizer.nextToken();

                int numQueries = Integer.parseInt(theTokenizer.nextToken());
                for (int i = 0; i < numQueries; i++) {
                    double thisPoint = Double.parseDouble(theTokenizer.nextToken());
                    theHeights.add(theState.getHeightAtPosition(thisPoint));
                }

                MCHeightResponse theResponseObject = new MCHeightResponse(theHeights);
                return theResponseObject.makeStringResponse();
            }

            if (theCustomType.startsWith("GETMCGOAL")) {
                MCGoalResponse theResponseObject = new MCGoalResponse(theState.goalPosition);
                return theResponseObject.makeStringResponse();
            }

        }*/
        System.err.println("We need some code written in Env Message for Playroom.. unknown request received: " + theMessage);
        Thread.dumpStack();
        return null;
    }

    public static void main(String[] args) {
        EnvironmentLoader L = new EnvironmentLoader(new Playroom());
        L.run();
    }

    /**
     * Turns theState object into an observation.
     * @return
     */
    @Override
    protected Observation makeObservation() {
        Observation currentObs = new Observation(7, 0);

        currentObs.intArray[0] = theState.whatIsAt(theState.EYE);
        currentObs.intArray[1] = theState.whatIsAt(theState.HAND);
        currentObs.intArray[2] = theState.whatIsAt(theState.MARKER);
        currentObs.intArray[3] = theState.isBellRinging?1:0;
        currentObs.intArray[4] = theState.isLightOn?1:0;
        currentObs.intArray[5] = theState.isMusicOn?1:0;
        currentObs.intArray[6] = theState.isMonkeyCrying?1:0;

        return currentObs;
    }

    public void env_cleanup() {
        if (savedStates != null) {
            savedStates.clear();
        }
    }

    /**
     * The value function will be drawn over the position and velocity.  This 
     * method provides the max values for those variables.
     * @param dimension
     * @return
     */
/*    public double getMaxValueForQuerableVariable(int dimension) {
        if (dimension == 0) {
            return theState.maxPosition;
        } else {
            return theState.maxVelocity;
        }
    }
*/
    /**
     * The value function will be drawn over the position and velocity.  This 
     * method provides the min values for those variables.
     * @param dimension
     * @return
     */
/*    public double getMinValueForQuerableVariable(int dimension) {
        if (dimension == 0) {
            return theState.minPosition;
        } else {
            return theState.minVelocity;
        }
    }
*/
    /**
     * Given a state, return an observation.  This is trivial in mountain car
     * because the observation is the same as the internal state 
     * @param theState
     * @return
     */
    public Observation getObservationForState(Observation theState) {
        return makeObservation();
    }

    /**
     * How many state variables are there (used for value function drawing)
     * @return
     */
/*    public int getNumVars() {
        return 2;
    }
*/
    /**
     * Used by MCHeightRequest Message
     * @return
     */
/*    private double getHeight() {
        return theState.getHeightAtPosition(theState.position);
    }

    public String getVisualizerClassName() {
        return PlayroomVisualizer.class.getName();
    }
*/
    private Random getRandomGenerator() {
        return randomGenerator;
    }

    /**
     * So we can draw a pretty image in the visualizer before we start
     * @return
     */
    public URL getImageURL() {
        URL imageURL = Playroom.class.getResource("/images/mountaincar.png");
        return imageURL;
    }
}

/**
 * This is a little helper class that fills in the details about this environment
 * for the fancy print outs in the visualizer application.
 * @author Nitish Srivastava
 */
class DetailsProvider implements hasVersionDetails {

    public String getName() {
        return "Playroom 1.0";
    }

    public String getShortName() {
        return "Playroom";
    }

    public String getAuthors() {
        return "Nitish Srivastava";
    }

    public String getInfoUrl() {
        return "http://home.iitk.ac.in/~nitishs/CS365/project";
    }

    public String getDescription() {
        return "RL-Library Java Version of Playroom.";
    }
}

