/**
* PlayroomState
*/
package org.rlcommunity.environments.mountaincar;
import java.util.Random;

/**
 * This class manages all of the problem parameters, current state variables, 
 * and state transition and reward dynamics.
 *
 * @author Nitish Srivastava
 */
public class PlayroomState {
//	Current State Information
    public static int numobjects=9;
    public boolean newgoal, oldgoal;
    public static final String[] Names={
	"EYE","HAND","MARKER","SWITCH",
	"BALL", "BELL", "MUSIC0", "MUSIC1", "MONKEY"};
    public int[][] position;
    public boolean isLightOn;
    public boolean isMusicOn;
    public boolean isMonkeyCrying;
    public boolean isBellRinging;
    public static int EYE=0;
    public static int HAND=1;
    public static int MARKER=2;
    public static int SWITCH=3;
    public static int BALL=4;
    public static int BELL=5;
    public static int MUSIC0=6;
    public static int MUSIC1=7;
    public static int MONKEY=8;

    public static int maxpos=3;
    public static int minpos=0;
    public static double rewardAtGoal= 10.0d;
    public static double rewardPerStep= -0.001d;
    private Random randomGenerator;

    //These are configurable
    public boolean randomStarts =true ;

    /**
     * Constructor 
     * @param randomGenerator
     */
    PlayroomState(Random randomGenerator) {
	position=new int[9][2];
	oldgoal=false;
	newgoal=false;
        this.randomGenerator = randomGenerator;
    }

    /**
     * Calculate the reward for the 
     * @return
     */
    public double getReward() {
        if (inGoalState()) {
            return rewardAtGoal;
        } else {
            return rewardPerStep;
        }
    }


    /**
     * IS the monkey crying?
     * @return
     */
    public boolean inGoalState() {
        return false;//(oldgoal && newgoal);
    }

   /**
    * update position. 
    * @param a Should be in [0,8]
    */
    void update(int a) {
	//a is the action code: change state accordingly
/*
1. move eye to hand : int
2) move eye to marker : int
3)+ move eye one step north, south, east or west : int[4]
4) move eye to random object : int
5) move hand to eye : int
6) move marker to eye. : int
7) do special action: int
In addition, if both the eye and and hand are on some object, then natural operations suggested
by the object become available, e.g., if both the hand and the eye are on the light switch,
then the action of flicking the light switch becomes available, and if both the hand and eye
are on the ball, then the action of kicking the ball becomes available (which when pushed,
moves in a straight line to the marker).
*/
	//System.out.println("Action received:"+a);
	if(atSamePos(HAND,EYE)){
		//System.out.println("Hand Eye same");
		if(atSamePos(HAND,SWITCH)){
			//System.out.println("Also Switch");
		//	isLightOn=true;
		}
	}
	newgoal=oldgoal;
	switch(a){
	    case 0:	position[EYE][0]=position[HAND][0];
			position[EYE][1]=position[HAND][1];
			break;
	    case 1:	position[EYE][0]=position[MARKER][0];
                        position[EYE][1]=position[MARKER][1];
                        break;
	    case 2:	position[EYE][0]=position[EYE][0]+1;
			if(position[EYE][0]>maxpos)position[EYE][0]=maxpos;
                        break;
	    case 3:     position[EYE][1]=position[EYE][1]+1;
                        if(position[EYE][1]>maxpos)position[EYE][1]=maxpos;
                        break;
	    case 4:     position[EYE][0]=position[EYE][0]-1;
                        if(position[EYE][0]<minpos)position[EYE][0]=minpos;
                        break;
	    case 5:     position[EYE][1]=position[EYE][1]-1;
                        if(position[EYE][1]<minpos)position[EYE][1]=minpos;
                        break;
	    case 6:	a=randomGenerator.nextInt(6)+3;
			//a=SWITCH;
			position[EYE][0]=position[a][0];
			position[EYE][1]=position[a][1];
			break;
	    case 7:	position[HAND][0]=position[EYE][0];
			position[HAND][1]=position[EYE][1];
			break;
	    case 8:	position[MARKER][0]=position[EYE][0];
                        position[MARKER][1]=position[EYE][1];
                        break;
	    case 9:	if(atSamePos(HAND,EYE)){
				//System.out.println("Came at same pos");
			    if(atSamePos(EYE, SWITCH)){
				System.out.println("Light switch used");
				isLightOn=!isLightOn;
				newgoal=true;
			    }
			    else if(atSamePos(EYE, BALL)){
				System.out.println("Kicking ball");
				position[BALL][0]=position[MARKER][0];
				position[BALL][1]=position[MARKER][1];
				if(atSamePos(BALL, BELL)){
				    System.out.println("Bell rung!");
				    isBellRinging=true;
				    int move=randomGenerator.nextInt(4);
				    if(move%2==0){
					position[BELL][0]+=move-1;
				    }
				    else{
					position[BELL][1]+=move-2;
				    }
				}
				if(!isLightOn && isMusicOn && isBellRinging){
				    isMonkeyCrying=true;
				}
			    }
			    else if(atSamePos(EYE, MUSIC0)){
				System.out.println("Turning music off");
				isMusicOn=false;
			    }
			    else if(atSamePos(EYE, MUSIC1)){
				System.out.println("Turning music on");
				isMusicOn=true;
			    } 
			}
			break;
	    default:
			System.err.println("Invalid action");
	}
    }
    

   public int whatIsAt(int object){
/*	for(int i=3;i<numobjects;i++){
		if(atSamePos(object, i))return i+1;
	}
*/	if(atSamePos(object, SWITCH))return 1;
	//if(atSamePos(object, MUSIC0))return 1;
	//if(atSamePos(object, MUSIC1))return 2;
	return 0;
   }
    public boolean atSamePos(int a, int b){
	return (position[a][0]==position[b][0] 
			&& position[a][1]==position[b][1]);
    }

    /**
     * This is basically a copy constructor and we use it when were doing 
     * env_save_state and env_save_state
     * @param stateToCopy
     */
    
    public PlayroomState(PlayroomState stateToCopy) {
        this.position = stateToCopy.position;
        this.isBellRinging=stateToCopy.isBellRinging;
	this.isMusicOn=stateToCopy.isMusicOn;
	this.isLightOn=stateToCopy.isLightOn;
	this.isMonkeyCrying=stateToCopy.isMonkeyCrying;
	this.rewardPerStep = stateToCopy.rewardPerStep;
        this.rewardAtGoal = stateToCopy.rewardAtGoal;
        this.randomStarts = stateToCopy.randomStarts;
        this.randomGenerator = stateToCopy.randomGenerator;
    }
    public void displayState(){
	System.out.println("\n***********Playroom State************");
	for(int i=0;i<numobjects;i++)
	System.out.println(Names[i]+" : "+position[i][0]+","+position[i][1]);

    }
}
